import json

from django import http, forms
from django.test.client import RequestFactory

from django_contracts.contracts import apply_contract
from unittest import mock


class MyUserForm(forms.Form):
    char = forms.CharField()

    def __init__(self, data, user, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        self.user = user
        if self.user.username == 'valid':
            self.fields['hidden'] = forms.CharField()


@apply_contract({'POST': MyUserForm}, {}, pass_in_user=True, )
def basic_view(request):
    return http.HttpResponse(
        content=json.dumps(request.validated_data),
        status=200,
    )


def test_user_a_cannot_use_hidden_field():
    user = mock.Mock(username='user_a')

    request = RequestFactory().post(
        path='/test/path/',
        data={
            'char': 'I am allowed to use this.',
            'hidden': 'Arrr! I will use any API I want!',
        }
    )

    request.user = user

    response = basic_view(request)

    response_dict = json.loads(response.content)

    assert response_dict == {
        'char': 'I am allowed to use this.',
    }

    assert response.status_code == 200


def test_user_a_cannot_use_hidden_field():
    user = mock.Mock(username='valid')

    request = RequestFactory().post(
        path='/test/path/',
        data={
            'char': 'I am allowed to use this.',
            'hidden': 'Huzzah! Great privilege.',
        }
    )

    request.user = user

    response = basic_view(request)

    response_dict = json.loads(response.content)

    assert response_dict == {
        'char': 'I am allowed to use this.',
        'hidden': 'Huzzah! Great privilege.',
    }

    assert response.status_code == 200


import json

from django import http, forms
from django.test.client import RequestFactory

from django_contracts.contracts import apply_contract


class MyAdvancedForm(forms.Form):
    char = forms.CharField()

    def clean_char(self):
        value = self.cleaned_data['char']

        if value == 'unauthorized choice':
            raise forms.ValidationError('Intruder!', code='unauthorized')

        elif value == 'missing choice':
            raise forms.ValidationError('404 not found.', code='not_found')

        elif value == 'custom choice':
            raise forms.ValidationError('Conflict!', code='custom')

        else:
            raise forms.ValidationError('Something else went wrong', code='invalid')

    class Meta:
        code_mapping = [
            ('custom', 409),
        ]


@apply_contract({'POST': MyAdvancedForm}, None, pass_in_user=False)
def my_advanced_view(request):
    return http.HttpResponse(
        content=json.dumps(request.validated_data),
        status=200,
    )


def test_my_advanced_view_returns_errors_unauthorized():
    request = RequestFactory().post(
        path='/test/path/',
        data={
            'char': 'unauthorized choice',
        }
    )

    response = my_advanced_view(request)

    response_dict = json.loads(response.content)

    assert response_dict == {
        'char': [
            {
                'code': 'unauthorized',
                'message': 'Intruder!',
            },
        ],
    }

    assert response.status_code == 403


def test_my_advanced_view_returns_errors_not_found():
    request = RequestFactory().post(
        path='/test/path/',
        data={
            'char': 'missing choice',
        }
    )

    response = my_advanced_view(request)

    response_dict = json.loads(response.content)

    assert response_dict == {
        'char': [
            {
                'code': 'not_found',
                'message': '404 not found.',
            },
        ],
    }

    assert response.status_code == 404


def test_my_advanced_view_returns_errors_invalid():
    request = RequestFactory().post(
        path='/test/path/',
        data={
            'char': 'invalid choice',
        }
    )

    response = my_advanced_view(request)

    response_dict = json.loads(response.content)

    assert response_dict == {
        'char': [
            {
                'code': 'invalid',
                'message': 'Something else went wrong',
            },
        ],
    }

    assert response.status_code == 400


def test_my_advanced_view_returns_errors_custom():
    request = RequestFactory().post(
        path='/test/path/',
        data={
            'char': 'custom choice',
        }
    )

    response = my_advanced_view(request)

    response_dict = json.loads(response.content)

    assert response_dict == {
        'char': [
            {
                'code': 'custom',
                'message': 'Conflict!',
            },
        ],
    }

    assert response.status_code == 409



import json

from django import http, forms


class ErrorForm(forms.Form):

    reason = forms.CharField()

    def clean(self):
        raise forms.ValidationError(
            [
                forms.ValidationError(
                    message=(
                        "Invalid POST data. "
                        "The supported MIME types for this endpoint are: "
                        "[multipart/form-data, application/json]"
                    ),
                    code="invalid"
                ),
                forms.ValidationError(
                    message=self.data['reason'],
                    code="invalid"
                ),
            ],
        )


def normalize_view_data(request):
    try:
        request.data = normalize_request_data(request, request.method)
    except json.JSONDecodeError as e:
        errors = ErrorForm({'reason': str(e)})
        errors.is_valid()
        return http.HttpResponseBadRequest(
            content=json.dumps(errors.errors),
            content_type="application/json"
        )

    return None


def normalize_request_data(request, request_type):
    body = request.body or '{}'

    # request.POST.dict() only takes a single value from array inputs so we use `lists` and transform
    # singular values back
    as_lists = getattr(request, request_type).lists()
    as_mix = {key: val[0] if val and len(val) == 1 else val for key, val in as_lists}

    # If there is no request.* try json
    normalized_data = as_mix or json.loads(body)

    return normalized_data



import json

from django import http, forms
from django.test.client import RequestFactory

from django_contracts.contracts import apply_contract


class MyTestForm(forms.Form):
    char = forms.CharField()


@apply_contract({'POST': MyTestForm}, {}, pass_in_user=True)
def basic_view(request):
    return http.HttpResponse(
        content=json.dumps(request.validated_data),
        status=200,
    )


def test_my_test_view_works():
    request = RequestFactory().post(
        path='/test/path/',
        data={
            'char': 'a lazy brown fox',
            'other': 'a cow jumps over the moon',
        }
    )

    response = basic_view(request)

    response_dict = json.loads(response.content)

    assert response_dict == {
        'char': 'a lazy brown fox'
    }

    assert response.status_code == 200


def test_my_test_view_returns_errors():
    request = RequestFactory().post(
        path='/test/path/',
        data={
            'char': '',
            'other': 'a cow jumps over the moon',
        }
    )

    response = basic_view(request)

    response_dict = json.loads(response.content)

    assert response_dict == {
        'char': [
            {
                'code': 'required',
                'message': 'This field is required.',
            },
        ],
    }

    assert response.status_code == 400


def test_bad_encoding_error():
    request = RequestFactory().post(
        path='/test/path/',
        body='some other formatting',
        encoding='application/json',
    )

    response = basic_view(request)
    response_dict = json.loads(response.content)

    assert response_dict == {
        '__all__': [
            (
                'Invalid POST data. The supported MIME types for this endpoint are: '
                '[multipart/form-data, application/json]'
            ),
            'Expecting value: line 1 column 1 (char 0)',
        ],
    }

    assert response.status_code == 400




from django import forms, http
from django_contracts.discovery.primitives import as_json, view_to_json

from django_contracts.contracts import apply_contract


class MyForm(forms.Form):
    char = forms.CharField(min_length=10, max_length=99, strip=True, empty_value='something', required=True)
    num = forms.IntegerField(min_value=-99, max_value=100, required=True, initial=100, )


@apply_contract(
    request_contracts={'GET': MyForm, 'POST': MyForm},
    response_contracts={'GET': MyForm, 'POST': MyForm},
    pass_in_user=True,
)
def basic_view(request):
    return http.HttpResponse(
        content='Hello World',
        status=200,
    )


def test_json_serializing_form_definition():
    json_repr = as_json(MyForm())

    assert json_repr == {
        'char': {
            'default': None,
            'required': True,
            'type': 'text',
            'maxlength': '99',
            'minlength': '10',
        },
        'num': {
            'default': 100,
            'required': True,
            'type': 'number',
            'min': -99,
            'max': 100,
        },
    }


def test_get_discovery_documentation_from_view_request_contract():
    json_repr = as_json(basic_view.contracts['request']['POST']())

    assert json_repr == {
        'char': {
            'default': None,
            'required': True,
            'type': 'text',
            'maxlength': '99',
            'minlength': '10',
        },
        'num': {
            'default': 100,
            'required': True,
            'type': 'number',
            'min': -99,
            'max': 100,
        },
    }


def test_get_discovery_documentation_from_view():
    json_repr = view_to_json(basic_view)

    assert json_repr == {
        'request': {
            'GET': {
                'char': {
                    'default': None,
                    'required': True,
                    'type': 'text',
                    'maxlength': '99',
                    'minlength': '10',
                },
                'num': {
                    'default': 100,
                    'required': True,
                    'type': 'number',
                    'min': -99,
                    'max': 100,
                },
            },
            'POST': {
                'char': {
                    'default': None,
                    'required': True,
                    'type': 'text',
                    'maxlength': '99',
                    'minlength': '10',
                },
                'num': {
                    'default': 100,
                    'required': True,
                    'type': 'number',
                    'min': -99,
                    'max': 100,
                },
            },
        },
        'response': {
            'GET': {
                'char': {
                    'default': None,
                    'required': True,
                    'type': 'text',
                    'maxlength': '99',
                    'minlength': '10',
                },
                'num': {
                    'default': 100,
                    'required': True,
                    'type': 'number',
                    'min': -99,
                    'max': 100,
                },
            },
            'POST': {
                'char': {
                    'default': None,
                    'required': True,
                    'type': 'text',
                    'maxlength': '99',
                    'minlength': '10',
                },
                'num': {
                    'default': 100,
                    'required': True,
                    'type': 'number',
                    'min': -99,
                    'max': 100,
                },
            },
        }
    }





def as_json(form):
    return {field_name: serialize_field(field) for field_name, field in form.fields.items()}


def serialize_field(field):
    return {
        'default': field.initial,
        'required': field.required,
        'type': field.widget.input_type,
        **field.widget_attrs(field.widget),
    }


def view_to_json(view):
    discovery_info = {}

    try:
        discovery_info['request'] = {key: as_json(value()) for key, value in view.contracts['request'].items()}
    except AttributeError:
        pass

    try:
        discovery_info['response'] = {key: as_json(value()) for key, value in view.contracts['response'].items()}
    except AttributeError:
        pass

    return discovery_info



from django import http


def handle_api_form_errors(form):
    response_status = 400  # Default to 400
    default_mapping = [
        ('unauthorized', 403),
        ('not_found', 404),
        ('missing', 404),
        ('invalid', 400),
    ]
    try:
        mapping = form.Meta.code_mapping
        full_mapping = mapping + default_mapping
    except AttributeError:
        full_mapping = default_mapping

    for code, status in full_mapping:
        if is_error_code_present(code, form.errors):
            response_status = status

    return http.HttpResponse(
        content_type='application/json',
        content=form.errors.as_json(),
        status=response_status,
    )


def is_error_code_present(error_code, form_errors):
    error_values = form_errors.as_data().values()
    return any(any(error.code == error_code for error in errors) for errors in error_values)



import logging
from functools import wraps
from .errors import handle_api_form_errors
from .normalizers import normalize_view_data

log = logging.getLogger(__name__)


def apply_contract(request_contracts=None, response_contracts=None, pass_in_user=False):
    def decorator(func):
        @wraps(func)
        def wrapper(request, *args, **kwargs):
            errors = normalize_view_data(request)
            if errors:
                return errors

            if response_contracts:

                method = request.method
                contract_for_method = response_contracts[method]

                if pass_in_user:
                    contract = contract_for_method(request.data, getattr(request, 'user', None))
                else:
                    contract = contract_for_method(request.data)

                if contract.errors:
                    log.info(msg=f"Request has errors: {contract.errors}")
                    return handle_api_form_errors(contract)

                try:
                    request.contracts['response'] = response_contracts
                except (AttributeError, KeyError):
                    request.contracts = {
                        'response': response_contracts
                    }

            if request_contracts:

                method = request.method
                contract_for_method = request_contracts[method]

                if pass_in_user:
                    contract = contract_for_method(request.data, getattr(request, 'user', None))
                else:
                    contract = contract_for_method(request.data)

                if contract.errors:
                    log.info(msg=f"Request has errors: {contract.errors}")
                    return handle_api_form_errors(contract)

                try:
                    request.contracts['request'] = contract
                except (AttributeError, KeyError):
                    request.contracts = {
                        'request': contract,
                    }

                request.validated_data = contract.cleaned_data

            return func(request, *args, **kwargs)

        try:
            wrapper.contracts['response'] = response_contracts
        except (AttributeError, KeyError):
            wrapper.contracts = {
                'response': response_contracts
            }

        try:
            wrapper.contracts['request'] = request_contracts
        except (AttributeError, KeyError):
            wrapper.contracts = {
                'request': request_contracts
            }
        return wrapper

    return decorator





def get_contract(request, request_contract, pass_in_user=False):
    if pass_in_user:
        return request_contract(request.data, request.user)
    else:
        return request_contract(request.data)
